
/**
 * Module dependencies.
 */

var EventEmitter = require('events').EventEmitter
  , packageTree = require('package-tree')
  , resolve = require('path').resolve
  , debug = require('debug')('component')
  , html = require('./plugins/html')
  , css = require('./plugins/css')
  , boot = require('./boot')
  , Batch = require('batch')
  , path = require('path')
  , extname = path.extname
  , join = path.join
  , fs = require('fs');

/**
 * Expose `Component`.
 */

module.exports = Component;

/**
 * Initialize a `Component` with the given `path`.
 *
 * @param {String} path
 * @api public
 */

function Component(path, tree) {
  this.root = ! tree;
  this.path = resolve(path);
  this._hooks = {};
  this.reset(tree);
  this.use(html());
  this.use(css());
}

/**
 * Inherits from `EventEmitter.prototype`.
 */

Component.prototype.__proto__ = EventEmitter.prototype;

/**
 * Filter components that become part of the build.
 *
 * @param {Function} fn
 * @api public
 */

Component.prototype.filter = function(fn){
  this._filter = fn;
};

/**
 * Use the given plugin `fn`.
 *
 * @param {Function} fn
 * @return {Component}
 * @api public
 */

Component.prototype.use = function(fn){
  fn(this);
  return this;
};

/**
 * Define or perform hook `name`.
 *
 * @param {String} name
 * @param {Mixed} fn
 * @return {Mixed}
 * @api private
 */

Component.prototype.hook = function(name, fn){
  if ('function' == typeof fn) return this.defineHook(name, fn);
  return this.performHook(name, fn);
};

/**
 * Perform hook `name` with `arg` reduced.
 *
 * @param {String} name
 * @return {Mixex}
 * @api priate
 */

Component.prototype.performHook = function(name, arg){
  var fns = this._hooks[name] || [];
  for (var i = 0, len = fns.length; i < len; ++i) {
    arg = fns[i](arg);
  }
  return arg;
};

/**
 * Define hook `name` with callback `fn()`.
 *
 * @param {String} name
 * @param {String} fn
 * @api private
 */

Component.prototype.defineHook = function(name, fn){
  (this._hooks[name] = this._hooks[name] || []).push(fn);
};

/**
 * Reset the component, reconstructing the tree
 * of dependencies for a re-build.
 *
 * @param {Object} tree
 * @api public
 */

Component.prototype.reset = function(tree){
  this.tree = tree || packageTree(this.path);
  this.tree.component = this.tree.component || {};
  this.rendered = {};
  this.templates = {};
  this.coerce();
  this.component = this.tree.component;
  this.deps = this.tree.deps;
  this.name = this.tree.name;
  this.emit('reset');
};

/**
 * Coerce the package tree into `Component`s.
 *
 * @api private
 */

Component.prototype.coerce = function(){
  var self = this;
  var deps = this.tree.deps;

  function coerce(pkg, i) {
    this[i] = pkg instanceof Component
      ? pkg
      : new Component(pkg.dirname + '/package.json', pkg);
    if (pkg.deps) pkg.deps.forEach(coerce, pkg.deps);
  }

  deps.forEach(coerce, deps);
};

/**
 * Resolve `path` relative to this `Component`'s dirname.
 *
 * @param {String} path
 * @return {String}
 * @api public
 */

Component.prototype.resolve = function(path){
  return resolve(this.tree.dirname, path);
};

/**
 * Read `path` relative to this `Component`'s dirname
 * and invoke `fn(err, str)`.
 *
 * @param {String} path
 * @param {Function} fn
 * @api public
 */

Component.prototype.read = function(path, fn){
  path = this.resolve(path);
  fs.readFile(path, 'utf8', fn);
};

/**
 * Require `path` relative to this `Component`'s dirname.
 *
 * @param {Type} name
 * @return {Type}
 * @api public
 */

Component.prototype.require = function(path){
  path = this.resolve(path);
  return require(path);
};

/**
 * Load components:
 *
 *  - load scripts
 *  - load styles
 *
 * @param {Function} fn
 * @api public
 */

Component.prototype.load = function(fn){
  var batch = new Batch
    , self = this;

  batch.push(function(done){
    debug('loading templates');
    self.loadTemplates(function(err){
      if (err) return done(err);
      debug('loading scripts');
      self.loadScripts(function(err){
        debug('loaded scripts');
        done(err);
      });
    });
  });

  batch.push(function(done){
    debug('loading styles');
    self.loadStyles(function(err){
      debug('loaded styles');
      done(err);
    });
  });

  batch.end(fn);
};

/**
 * Load and compile templates and
 * re-assign their `component.templates[NAME]` value.
 *
 * @param {Function} fn
 * @api private
 */

Component.prototype.loadTemplates = function(fn){
  var batch = new Batch
    , self = this;

  this.traverse(function(pkg){
    var conf = pkg.tree.component;
    if (!conf.templates) return;
    Object.keys(conf.templates).forEach(function(name){
      var path = pkg.resolve(conf.templates[name]);
      batch.push(function(done){
        debug('%s : loading template %s %s', pkg.name, name, path);
        pkg.read(path, function(err, str){
          if (err) return done(err);
          var ext = conf.engine || extname(path).slice(1);
          self.emit('template:' + ext, path, str, function(err, fn){
            if (err) return done(err);
            if (!fn) return done(new Error('failed to compile template "' + path + '"'));
            conf._templates = conf._templates || {};
            conf._templates[name] = fn;
            debug('%s : loaded template %s', pkg.name, name);
            done();
          });
        });
      });
    });
  });

  batch.end(fn);
};

/**
 * Load scripts and re-assign
 * their `component.scripts[NAME]` value.
 *
 * @param {Function} fn
 * @api private
 */

Component.prototype.loadScripts = function(fn){
  var batch = new Batch
    , self = this;

  this.traverse(function(pkg){
    var conf = pkg.tree.component;
    if (!conf.scripts) return;
    Object.keys(conf.scripts).forEach(function(name){
      var path = pkg.resolve(conf.scripts[name]);
      batch.push(function(done){
        debug('%s : loading script %s %s', pkg.name, name, path);
        pkg.read(path, function(err, str){
          if (err) return done(err);
          debug('%s : loaded script %s', pkg.name, name);
          conf._scripts = conf._scripts || {};
          conf._scripts[name] = str;
          done();
        });
      });
    });
  });

  batch.end(fn);
};

/**
 * Load styles and re-assign
 * their `component.styles[i]` value.
 *
 * @param {Function} fn
 * @api private
 */

Component.prototype.loadStyles = function(fn){
  var batch = new Batch
    , self = this;

  this.reverseTraverse(function(pkg){
    var conf = pkg.tree.component;
    if (!conf.styles) return;
    conf.styles.forEach(function(name, i){
      var path = pkg.resolve(name);
      batch.push(function(done){
        debug('%s : loading style %s %s', pkg.name, name, path);
        pkg.read(path, function(err, str){
          if (err) return done(err);
          debug('%s : loaded style %s', pkg.name, name);
          var ext = extname(path).slice(1);
          self.emit('stylesheet:' + ext, path, str, function(err, css){
            if (err) return done(err);
            conf._styles = conf._styles || [];
            conf._styles[i] = css || '';
            done();
          });
        });
      });
    });
  });

  batch.end(fn);
};

/**
 * Aggregate scripts.
 *
 * @return {String}
 * @api private
 */

Component.prototype.aggregateScripts = function(){
  var self = this;
  var tmpls = [];
  var js = [];

  this.traverse(function(pkg){
    var conf = pkg.tree.component;
    if (!conf._scripts) return;
    debug('%s : aggregate scripts', pkg.name);
    Object.keys(conf._scripts).forEach(function(name){
      var str = '// component ' + pkg.name + ': ' + name + '\n';
      str += 'require.register("' + name + '", function(module, exports, require){\n';
      if (conf._templates) {
        tmpls.push(pkg.aggregateTemplates(self.rendered, pkg, conf));
        str += render.toString().replace('NAME', pkg.name) + '\n';
      }
      str += conf._scripts[name] + '\n';
      str += '});\n';
      js.push(str);
    });
  })

  return tmpls.concat(js).join('\n');
};

/**
 * Aggregate styles.
 *
 * @return {String}
 * @api private
 */

Component.prototype.aggregateStyles = function(){
  var css = [];

  this.traverse(function(pkg){
    var conf = pkg.tree.component;
    if (!conf._styles) return;
    debug('%s : aggregate styles', pkg.name);
    css = css.concat(conf._styles);
  })

  return css.join('\n');
};

/**
 * Aggregate and prepare templates
 * with file-scope. For example
 * an "album" component may have
 * self-contained templates, available
 * only to it.
 *
 * @param {Object} tmpls
 * @return {String}
 * @api private
 */

Component.prototype.aggregateTemplates = function(rendered, pkg, conf){
  var self = this;
  var tmpls = conf._templates;
  return Object.keys(tmpls).map(function(name){
    var key = pkg.name + ':' + name;
    if (rendered[key]) return;
    rendered[key] = true;
    var fn = tmpls[name];
    fn = fn.toString().replace('anonymous', '');
    return '__templates["' + key + '"] = ' + fn;
  }).join('\n') + '\n';
};

/**
 * Build scripts, outputting to `path`
 * and callback `fn(err)`. When `bootstrap` is
 * true the component scripts themselves are 
 * aggregated.
 *
 * @param {String} path
 * @param {Function} fn
 * @param {Boolean} bootstrap
 * @api private
 */

Component.prototype.buildScripts = function(path, fn, bootstrap){
  var self = this
    , bootstrap = false === bootstrap ? false : true;

  // TODO: this.loadScripts only

  debug('building to %s', path);
  this.load(function(err){
    if (err) return fn(err);
    boot.scripts(function(err, boot){
      if (err) return fn(err);
      var js = bootstrap
        ? boot + self.aggregateScripts()
        : self.aggregateScripts();

      if (self.listeners('scripts').length) {
        self.emit('scripts', js, function(err, js){
          if (err) return fn(err);
          fs.writeFile(path + '.js', js, fn);
        });
      } else {
        fs.writeFile(path + '.js', js, fn);
      }
    });
  });
};

/**
 * Build styles, outputting to `path`
 * and callback `fn(err)`.
 *
 * @param {String} path
 * @param {Function} fn
 * @api private
 */

Component.prototype.buildStyles = function(path, fn, bootstrap){
  var self = this
    , bootstrap = false === bootstrap ? false : true;

  // TODO: this.loadStyles only

  debug('building to %s', path);
  this.load(function(err){
    if (err) return fn(err);
    var css = self.aggregateStyles();
    if (self.listeners('styles').length) {
      self.emit('styles', css, function(err, css){
        if (err) return fn(err);
        fs.writeFile(path + '.css', css, fn);
      });
    } else {
      fs.writeFile(path + '.css', css, fn);
    }
  });
};

/**
 * Build files to `path` and invoke `fn(err)`.
 *
 * For example "/tmp/foo" will produce
 * "/tmp/foo.css" and "/tmp/foo.js".
 *
 * TODO: variants that just reply with strings...
 *
 * @param {String} path
 * @param {Function} fn
 * @api public
 */

Component.prototype.build = function(path, fn, bootstrap){
  var self = this
    , batch = new Batch
    , bootstrap = false === bootstrap ? false : true;

  debug('building to %s', path);
  this.load(function(err){
    if (err) return fn(err);
    batch.push(function(done){
      boot.scripts(function(err, boot){
        if (err) return done(err);
        // TODO: less of a hack
        var js = bootstrap
          ? boot + self.aggregateScripts()
          : self.aggregateScripts();

        js = self.hook('scripts', js);
        fs.writeFile(path + '.js', js, done);
      });
    });

    batch.push(function(done){
      var css = self.aggregateStyles();
      css = self.hook('styles', css);
      fs.writeFile(path + '.css', css, done);
    });

    batch.end(fn);
  });
};

/**
 * Visit components and invoke `fn(component)`
 * from the top down.
 *
 * @param {Function} fn
 * @param {Function} fn2
 * @api public
 */

Component.prototype.traverse = function(fn, fn2, hit, filter){
  hit = hit || {};
  if (this._filter) filter = this._filter;
  if (hit[this.name]) return;
  if (filter && !filter(this)) return;
  hit[this.name] = true;

  fn(this);
  this.deps.forEach(function(dep){
    dep.traverse(fn, fn2, hit, filter);
  });
  fn2 && fn2(this);
};

/**
 * Visit components and invoke `fn(component)`
 * from the bottom up.
 *
 * @param {Function} fn
 * @api public
 */

Component.prototype.reverseTraverse = function(fn, hit){
  hit = hit || {};
  if (hit[this.name]) return;
  hit[this.name] = true;

  this.deps.forEach(function(dep){
    dep.reverseTraverse(fn, hit);
  });
  fn(this);
};

/**
 * Render template.
 */

function render(view, options) {
  return __templates['NAME:' + view](options);
}
