
/**
 * Load one or more deps.
 *
 * @param {Array|String} dep(s)
 * @api public
 */

function load(deps, fn) {
  if ('string' == typeof deps) {
    load.dep(deps, fn);
  } else {
    fn = load.pending(deps.length, fn);
    deps.forEach(function(dep){
      load.dep(dep, fn);
    });
  }
}

/**
 * Invoke `fn()` when dom content
 * is loaded or previously complete.
 * 
 * @param {Function} fn
 * @api private
 */

load.domready = function(fn){
  if (load.complete) return fn();
  document.addEventListener('DOMContentLoaded', fn, false);
};

/**
 * Load the given `dep` and it's entire
 * dependency tree.
 *
 * @param {String} dep
 * @param {Function} fn
 * @api public
 */

load.dep = function(dep, fn){
  fn = load.pending(2, fn);
  load.style('/component/' + dep + '.css', fn);
  load.domready(function(){
    load.script('/component/' + dep + '.js', fn);
  });
};

/**
 * Return a function which invokes `fn(err)`
 * on completetion of `n` calls.
 *
 * @param {Number} n
 * @param {Function} fn
 * @return {Function}
 * @api private
 */

load.pending = function(n, fn) {
  var done;
  return function(err){
    if (done) return;
    if (err) return done = true, fn(err);
    --n || fn();
  }
};

/**
 * Load style `path` and invoke `fn()`.
 *
 * @param {String} path
 * @param {Function} fn
 * @api private
 */

load.style = function(path, fn){
  var el = document.createElement('link');
  el.setAttribute('rel', 'stylesheet');
  el.setAttribute('href', path);
  el.setAttribute('type', 'text/css');
  document.head.appendChild(el);
  el.onload = function(){ fn(); };
};

/**
 * Load script `path` and invoke `fn()`.
 *
 * @param {String} path
 * @param {Function} fn
 * @api private
 */

load.script = function(path, fn){
  var el = document.createElement('script');
  el.setAttribute('async', 'true');
  el.src = path;
  document.body.appendChild(el);
  el.onload = function(){ fn(); };
};

// dom loaded flag

load.domready(function(){ load.complete = true; });